@@ -1,4 +1,5 @@
 use crate::codegen::contextual::Contextual;
+use crate::codegen::named::FnOverloadName;
 use crate::codegen::named::{CasedTypeIdent, Named};
 use crate::codegen::resolve_target_type::ResolveTargetType;
 use crate::codegen::serialization_type::{SerializationType, SerializationTypeGetter};
@@ -392,7 +393,7 @@ pub mod fn_types {
         let typ = with_context(typ, in_context);
         match serialization_type {
             SerializationType::Raw | SerializationType::JsValue => quote! { #typ },
-            SerializationType::SerdeJson => quote! { JsValue },
+            SerializationType::SerdeJson | SerializationType::Array => quote! { JsValue },
             SerializationType::Fn => {
                 let target = typ.resolve_target_type();
                 match target {
@@ -455,7 +456,10 @@ pub mod fn_types {
         let typ = with_context(typ, in_context);
         let typ = typ.as_ref();
         match serialization_type {
-            SerializationType::Raw | SerializationType::SerdeJson | SerializationType::JsValue => {
+            SerializationType::Raw
+            | SerializationType::SerdeJson
+            | SerializationType::JsValue
+            | SerializationType::Array => {
                 quote! { #typ }
             }
             SerializationType::Fn => match typ {
@@ -772,7 +776,7 @@ impl<T: HasFnPrototype + ?Sized> FnPrototypeExt for T {
         let f = self.exposed_to_rust_fn_decl(fn_name, true, in_context);
 
         quote! {
-            #[allow(dead_code)]
+            #[allow(dead_code, non_snake_case)]
             pub #f {
                 #wrapper_fns
 
@@ -843,7 +847,7 @@ impl<T: HasFnPrototype + ?Sized> FnPrototypeExt for T {
         let f = self.exposed_to_rust_fn_decl(fn_name, is_fallible, in_context);
 
         quote! {
-            #[allow(dead_code)]
+            #[allow(dead_code, non_snake_case)]
             pub #f {
                 #(#arg_converters)*
 
@@ -1065,7 +1069,7 @@ impl<T: WrappedParam> ParamExt for T {
 
         match serialization_type {
             SerializationType::Raw | SerializationType::JsValue => quote! { #name },
-            SerializationType::SerdeJson => {
+            SerializationType::SerdeJson | SerializationType::Array => {
                 quote! {
                     ts_bindgen_rt::from_jsvalue(&#name).map_err(ts_bindgen_rt::Error::from)?
                 }
@@ -1289,8 +1293,8 @@ pub struct InternalFunc<'a> {
 }
 
 impl<'a> InternalFunc<'a> {
-    pub fn to_internal_rust_name(js_name: &str) -> Identifier {
-        Self::to_internal_rust_ident(&to_snake_case_ident(js_name))
+    pub fn to_internal_rust_name(&self) -> Identifier {
+        Self::to_internal_rust_ident(&self.func.overload_name(&to_snake_case_ident(self.js_name)))
     }
 
     pub fn to_internal_rust_ident(id: &Identifier) -> Identifier {
@@ -1300,7 +1304,7 @@ impl<'a> InternalFunc<'a> {
 
 impl<'a> ToTokens for InternalFunc<'a> {
     fn to_tokens(&self, toks: &mut TokenStream2) {
-        let fn_name = Self::to_internal_rust_name(self.js_name);
+        let fn_name = self.to_internal_rust_name();
 
         let f = self.func.exposed_to_js_fn_decl(fn_name, *self.in_context);
 
@@ -1320,7 +1324,7 @@ fn render_wasm_bindgen_return_to_js(
     let serialization_type = return_type.serialization_type();
     match serialization_type {
         SerializationType::Raw | SerializationType::JsValue => return_value.clone(),
-        SerializationType::SerdeJson => {
+        SerializationType::SerdeJson | SerializationType::Array => {
             if is_fallible {
                 quote! {
                     ts_bindgen_rt::from_jsvalue(&#return_value?).map_err(ts_bindgen_rt::Error::from).map_err(JsValue::from)
@@ -1353,7 +1357,7 @@ pub fn render_raw_return_to_js(return_type: &TypeRef, return_value: &TokenStream
             #return_value.into_serde().unwrap()
         },
         SerializationType::JsValue => return_value.clone(),
-        SerializationType::SerdeJson => {
+        SerializationType::SerdeJson | SerializationType::Array => {
             quote! {
                 ts_bindgen_rt::from_jsvalue(&#return_value).unwrap()
             }
@@ -1383,6 +1387,51 @@ fn render_rust_to_js_conversion(
             quote! { ts_bindgen_rt::to_jsvalue(&#name)#error_mapper },
         ),
         SerializationType::Fn => (false, quote! { &#fn_name }),
+        SerializationType::Array => {
+            let tr = typ.as_ref();
+            let inner_type = if let TypeIdent::Builtin(Builtin::Array) = &tr.referent {
+                tr.type_params
+                    .first()
+                    .map(Cow::Borrowed)
+                    .unwrap_or_else(|| {
+                        Cow::Owned(TypeRef {
+                            referent: TypeIdent::Builtin(Builtin::PrimitiveAny),
+                            type_params: Default::default(),
+                            context: tr.context.clone(),
+                        })
+                    })
+            } else {
+                // TODO: this is an error
+                Cow::Owned(TypeRef {
+                    referent: TypeIdent::Builtin(Builtin::PrimitiveAny),
+                    type_params: Default::default(),
+                    context: tr.context.clone(),
+                })
+            };
+            let inner_name = name.suffix_name("_item");
+            let inner_conversion = render_rust_to_jsvalue_conversion(
+                &inner_name,
+                fn_name, // TODO: what do we do here?
+                &typ.similarly_wrap(inner_type.as_ref()),
+                quote! { ? },
+            );
+            (
+                true,
+                quote! {
+                    #name.into_iter().fold(
+                        Ok(js_sys::Array::new()) as std::result::Result<js_sys::Array, JsValue>,
+                        |arr, #inner_name| {
+                            #[allow(clippy::bind_instead_of_map)]
+                            arr.and_then(|a| {
+                                a.push(&#inner_conversion);
+                                Ok(a)
+                            })
+                        },
+                    )#error_mapper
+                    .into()
+                },
+            )
+        }
     }
 }
 
@@ -1392,34 +1441,45 @@ fn render_rust_to_jsvalue_conversion(
     typ: &TypeRefLike,
     error_mapper: TokenStream2,
 ) -> TokenStream2 {
+    let (_, conv) = render_rust_to_js_conversion(name, fn_name, typ, error_mapper);
     let serialization_type = typ.serialization_type();
-    match serialization_type {
-        SerializationType::Raw | SerializationType::JsValue => quote! { JsValue::from(#name) },
-        SerializationType::SerdeJson => {
-            quote! { ts_bindgen_rt::to_jsvalue(&#name)#error_mapper }
-        }
-        SerializationType::Fn => {
-            quote! { &#fn_name }
-        }
+    if matches!(
+        serialization_type,
+        SerializationType::Raw | SerializationType::JsValue
+    ) {
+        quote! { JsValue::from(#name) }
+    } else {
+        conv
     }
 }
 
 pub struct WrapperFunc<'a> {
     pub func: &'a Func,
     pub js_name: &'a str,
+    pub is_overloaded: bool,
 }
 
 impl<'a> WrapperFunc<'a> {
-    fn to_rust_name(js_name: &str) -> Identifier {
-        to_snake_case_ident(js_name)
+    fn to_rust_name(&self) -> Identifier {
+        let base_name = to_snake_case_ident(self.js_name);
+        if self.is_overloaded {
+            self.func.overload_name(&base_name)
+        } else {
+            base_name
+        }
     }
 }
 
 impl<'a> ToTokens for WrapperFunc<'a> {
     fn to_tokens(&self, toks: &mut TokenStream2) {
-        let fn_name = Self::to_rust_name(self.js_name);
+        let fn_name = self.to_rust_name();
 
-        let internal_fn_name = InternalFunc::to_internal_rust_name(self.js_name);
+        let internal = InternalFunc {
+            func: self.func,
+            js_name: self.js_name,
+            in_context: &None, // TODO: abstraction leak - we should have in_context
+        };
+        let internal_fn_name = internal.to_internal_rust_name();
         let our_toks = self
             .func
             .exposed_to_rust_wrapper_fn(&fn_name, &internal_fn_name, None);
