@@ -16,7 +16,7 @@ use crate::codegen::generics::{
     apply_type_params, render_type_params, render_type_params_with_constraints,
     render_type_params_with_lifetimes, ResolveGeneric, TypeEnvImplying,
 };
-use crate::codegen::named::{CasedTypeIdent, Named, SimpleNamed};
+use crate::codegen::named::{type_name, CasedTypeIdent, FnOverloadName, Named, SimpleNamed};
 use crate::codegen::ns_path::ToNsPath;
 use crate::codegen::resolve_target_type::ResolveTargetType;
 use crate::codegen::serialization_type::{SerializationType, SerializationTypeGetter};
@@ -24,13 +24,12 @@ use crate::codegen::traits::{render_trait_defn, to_type_ref, IsTraitable, TraitN
 use crate::codegen::type_ref_like::OwnedTypeRef;
 use crate::fs::Fs;
 use crate::identifier::{
-    make_identifier, to_camel_case_ident, to_ident, to_snake_case_ident, to_unique_ident,
-    Identifier,
+    to_camel_case_ident, to_ident, to_snake_case_ident, to_unique_ident, Identifier,
 };
 use crate::ir::{
-    Alias, Builtin, Class, Context, Enum, EnumMember, EnumValue, Func, Indexer, Interface,
-    Intersection, Member, NamespaceImport, TargetEnrichedType, TargetEnrichedTypeInfo, Tuple,
-    TypeIdent, TypeParamConfig, TypeRef, Union,
+    Alias, Builtin, Class, Context, Ctor, Enum, EnumMember, EnumValue, Func, Indexer, Interface,
+    Intersection, Member, NamespaceImport, Param, TargetEnrichedType, TargetEnrichedTypeInfo,
+    Tuple, TypeIdent, TypeParamConfig, TypeRef, Union,
 };
 pub use crate::mod_def::ModDef;
 use proc_macro2::TokenStream as TokenStream2;
@@ -240,14 +239,7 @@ impl IsUninhabited for TypeRef {
 }
 
 fn type_to_union_case_name(typ: &TargetEnrichedTypeInfo) -> Identifier {
-    let t_str = quote! { #typ }
-        .to_string()
-        .replace("<", "Of")
-        .replace(">", "")
-        .replace("&", "")
-        .replace("[", "")
-        .replace("]", "");
-    to_camel_case_ident(format!("{}Case", t_str))
+    type_name(typ)
 }
 
 fn path_relative_to_cargo_toml<T: AsRef<Path>>(path: T) -> PathBuf {
@@ -317,7 +309,7 @@ impl FieldCountGetter for TargetEnrichedTypeInfo {
             },
             TargetEnrichedTypeInfo::Tuple(t) => t.types.len(),
             TargetEnrichedTypeInfo::Mapped { .. } => usize::MAX,
-            TargetEnrichedTypeInfo::Func(_) => min,
+            TargetEnrichedTypeInfo::FuncGroup(_) => min,
             TargetEnrichedTypeInfo::Constructor(_) => min,
             TargetEnrichedTypeInfo::Class(c) => {
                 c.members.len()
@@ -611,8 +603,8 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                     .iter()
                     .filter_map(|t| {
                         let case = type_to_union_case_name(t).to_snake_case();
-                        let serialize_fn = render_serialize_fn(&case, &t);
-                        let deserialize_fn = render_deserialize_fn(&case, &t);
+                        let serialize_fn = render_serialize_fn(&case, t);
+                        let deserialize_fn = render_deserialize_fn(&case, t);
                         serialize_fn.zip(deserialize_fn).map(|(s, d)| {
                             quote! {
                                 #s
@@ -647,31 +639,43 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                     pub struct #name(#(pub #types),*);
                 }
             }
-            TargetEnrichedTypeInfo::Func(func) => {
-                let path = func.context.js_module_path();
-                let attrs = {
-                    let mut attrs = vec![quote! { js_name = #js_name, catch }];
-                    if func.is_variadic() {
-                        attrs.push(quote! { variadic });
+            TargetEnrichedTypeInfo::FuncGroup(func_group) => {
+                let path = func_group.context.js_module_path();
+                let common_attrs = vec![quote! { js_name = #js_name, catch }];
+                let is_overloaded = func_group.overloads.len() > 1;
+                let funcs = func_group.overloads.iter().map(|func| {
+                    let attrs = {
+                        let mut attrs = common_attrs.clone();
+                        if func.is_variadic() {
+                            attrs.push(quote! { variadic });
+                        }
+                        attrs
+                    };
+                    let internal_func = InternalFunc {
+                        js_name,
+                        func,
+                        in_context: &None,
+                    };
+                    let wrapper_func = WrapperFunc {
+                        js_name,
+                        func,
+                        is_overloaded,
+                    };
+
+                    quote! {
+                        #[wasm_bindgen(module=#path)]
+                        extern "C" {
+                            #[allow(non_snake_case)]
+                            #[wasm_bindgen(#(#attrs),*)]
+                            #internal_func
+                        }
+
+                        #wrapper_func
                     }
-                    attrs
-                };
-                let internal_func = InternalFunc {
-                    js_name,
-                    func,
-                    in_context: &None,
-                };
-                let wrapper_func = WrapperFunc { js_name, func };
+                });
 
                 quote! {
-                    #[wasm_bindgen(module=#path)]
-                    extern "C" {
-                        #[allow(non_snake_case)]
-                        #[wasm_bindgen(#(#attrs),*)]
-                        #internal_func
-                    }
-
-                    #wrapper_func
+                    #(#funcs)*
                 }
             }
             TargetEnrichedTypeInfo::Class(class) => {
@@ -784,77 +788,143 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                     .collect();
 
                 let target = quote! { self.0 };
-                let (member_defs, public_methods): (Vec<TokenStream2>, Vec<TokenStream2>) = members.iter()
-                    .map(|(member_js_name, member)| {
-                        let member_js_ident = format_ident!("{}", member_js_name);
-                        let internal_fn_name = InternalFunc::to_internal_rust_name(member_js_name);
-                        match member {
-                            Member::Constructor(ctor) => {
-                                let ctor = ctor.resolve_generic_in_env(&type_env);
-                                let ctor = Constructor::new(
-                                    ctor,
+                let ctor_name = |overloads: &[Ctor], ctor: &Constructor| {
+                    let is_overloaded = overloads.len() > 1;
+                    let fn_group_name = to_snake_case_ident("new");
+                    if is_overloaded {
+                        ctor.overload_name(&fn_group_name)
+                    } else {
+                        fn_group_name
+                    }
+                };
+                let default_ctor_name = members.iter().find_map(|(_, member)| {
+                    if let Member::Constructor(ctor) = member {
+                        ctor.overloads
+                            .iter()
+                            .find(|o| o.params.is_empty())
+                            .map(|o| {
+                                let c = Constructor::new(
+                                    Cow::Borrowed(o),
                                     TypeIdent::LocalName(js_name.to_string()),
                                 );
-                                let param_toks = ctor
-                                    .params()
-                                    .map(|p| p.as_exposed_to_js_named_param_list(None));
+                                ctor_name(&ctor.overloads, &c)
+                            })
+                    } else {
+                        None
+                    }
+                });
+                let default_impl = default_ctor_name
+                    .map(|default_ctor_name| {
+                        quote! {
+                            impl #full_type_params std::default::Default for #name #full_type_params {
+                                fn default() -> Self {
+                                    Self::#default_ctor_name()
+                                }
+                            }
+                        }
+                    })
+                    .unwrap_or_default();
+                let (member_defs, public_methods): (Vec<TokenStream2>, Vec<TokenStream2>) = members.iter()
+                    .flat_map(|(member_js_name, member)| {
+                        let member_js_ident = format_ident!("{}", member_js_name);
+                        match member {
+                            Member::Constructor(ctor) => {
+                                let overloads = &ctor.overloads;
+                                overloads
+                                    .iter()
+                                    .map(|ctor| {
+                                        let ctor = ctor.resolve_generic_in_env(&type_env);
+                                        let ctor = Constructor::new(
+                                            ctor,
+                                            TypeIdent::LocalName(js_name.to_string()),
+                                        );
+                                        let param_toks = ctor
+                                            .params()
+                                            .map(|p| p.as_exposed_to_js_named_param_list(None));
 
-                                let member_def = quote! {
-                                    #[wasm_bindgen(constructor, js_class = #js_name)]
-                                    pub fn new(#(#param_toks),*) -> #internal_class_name;
-                                };
-                                let fq_internal_ctor = to_snake_case_ident("new").in_namespace(&internal_class_name);
+                                        let fn_name = ctor_name(overloads, &ctor);
 
-                                let res_converter = |res: TokenStream2| -> TokenStream2 {
-                                    let args = if type_params.is_empty() {
-                                        vec![quote! { #res }]
-                                    } else {
-                                        vec![
-                                            quote! { #res },
-                                            quote! { std::marker::PhantomData #full_type_params },
-                                        ]
-                                    };
-                                    quote! {
-                                        #name(#(#args),*)
-                                    }
-                                };
-                                let pub_fn = ctor.exposed_to_rust_generic_wrapper_fn(&make_identifier!(new), None, &fq_internal_ctor, false, Some(&res_converter), &type_env, None);
+                                        let member_def = quote! {
+                                            #[wasm_bindgen(constructor, js_class = #js_name)]
+                                            pub fn #fn_name(#(#param_toks),*) -> #internal_class_name;
+                                        };
+                                        let fq_internal_ctor = fn_name.in_namespace(&internal_class_name);
 
-                                (member_def, pub_fn)
+                                        let res_converter = |res: TokenStream2| -> TokenStream2 {
+                                            let args = if type_params.is_empty() {
+                                                vec![quote! { #res }]
+                                            } else {
+                                                vec![
+                                                    quote! { #res },
+                                                    quote! { std::marker::PhantomData #full_type_params },
+                                                ]
+                                            };
+                                            quote! {
+                                                #name(#(#args),*)
+                                            }
+                                        };
+                                        let pub_fn = ctor.exposed_to_rust_generic_wrapper_fn(&fn_name, None, &fq_internal_ctor, false, Some(&res_converter), &type_env, None);
+
+                                        (member_def, pub_fn)
+                                    })
+                                    .collect()
                             }
                             Member::Method(func) => {
-                                let func = {
-                                    // func.class_name refers to our wrapper, which is what we want
-                                    // other than when we're defining our actual methods
-                                    let mut func = func.clone();
-                                    func.class_name = func.class_name.map(|_| TypeIdent::ExactName(internal_class_name.to_string()));
-                                    func
-                                };
-                                let func = func.resolve_generic_in_env(&type_env);
-                                let fn_name = InternalFunc::to_internal_rust_name(member_js_name);
+                                let is_overloaded = func.overloads.len() > 1;
+                                func.overloads
+                                    .iter()
+                                    .map(|func| {
+                                        let func = {
+                                            let mut func = func.clone();
+                                            func.class_name = func.class_name.map(|_| TypeIdent::ExactName(internal_class_name.to_string()));
+                                            func
+                                        };
+                                        let func = func.resolve_generic_in_env(&type_env);
+                                        let in_context = None;
+                                        let internal = InternalFunc {
+                                            func: &func,
+                                            js_name: member_js_name,
+                                            in_context: &in_context,
+                                        };
+                                        let fn_name = internal.to_internal_rust_name();
 
-                                let f = func.exposed_to_js_fn_decl(fn_name, None);
+                                        let f = func.exposed_to_js_fn_decl(fn_name, in_context);
 
-                                let mut attrs = vec![
-                                    quote! {js_name = #member_js_ident},
-                                    quote! {method},
-                                    quote! {js_class = #js_name},
-                                    quote! {catch},
-                                ];
-                                if func.is_variadic() {
-                                    attrs.push(quote! { variadic });
-                                }
+                                        let mut attrs = vec![
+                                            quote! {js_name = #member_js_ident},
+                                            quote! {method},
+                                            quote! {js_class = #js_name},
+                                            quote! {catch},
+                                        ];
+                                        if func.is_variadic() {
+                                            attrs.push(quote! { variadic });
+                                        }
 
-                                let member_def = quote! {
-                                    #[allow(non_snake_case)]
-                                    #[wasm_bindgen(#(#attrs),*)]
-                                    #f;
-                                };
+                                        let member_def = quote! {
+                                            #[allow(non_snake_case)]
+                                            #[wasm_bindgen(#(#attrs),*)]
+                                            #f;
+                                        };
 
-                                let rc: Option<&fn(TokenStream2) -> TokenStream2> = None;
-                                let pub_fn = func.exposed_to_rust_generic_wrapper_fn(&to_snake_case_ident(&member_js_name), Some(&target), &internal_fn_name, true, rc, &type_env, None);
+                                        let rc: Option<&fn(TokenStream2) -> TokenStream2> = None;
+                                        let in_context = None;
+                                        let internal = InternalFunc {
+                                            func: func.as_ref(),
+                                            js_name: member_js_name,
+                                            in_context: &in_context,
+                                        };
+                                        let internal_fn_name = internal.to_internal_rust_name();
+                                        let fn_group_name = to_snake_case_ident(&member_js_name);
+                                        let fn_name = if is_overloaded {
+                                            func.overload_name(&fn_group_name)
+                                        } else {
+                                            fn_group_name
+                                        };
+                                        let pub_fn = func.exposed_to_rust_generic_wrapper_fn(&fn_name, Some(&target), &internal_fn_name, true, rc, &type_env, in_context);
 
-                                (member_def, pub_fn)
+                                        (member_def, pub_fn)
+                                    })
+                                    .collect()
                             }
                             Member::Property(typ) => {
                                 let resolved_type = typ.resolve_generic_in_env(&type_env).into_owned();
@@ -919,7 +989,7 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                                     #setter_fn
                                 };
 
-                                (member_def, pub_fn)
+                                vec![(member_def, pub_fn)]
                             }
                         }
                     })
@@ -937,6 +1007,7 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                 quote! {
                     #[wasm_bindgen(module = #path)]
                     extern "C" {
+                        #[allow(non_camel_case_types)]
                         #[wasm_bindgen(#(#attrs),*)]
                         #vis type #internal_class_name;
 
@@ -948,6 +1019,34 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
 
                     #(#super_as_ref_impls)*
 
+                    #default_impl
+
+                    impl #full_type_params std::convert::From<#name #full_type_params> for JsValue {
+                        fn from(src: #name #full_type_params) -> JsValue {
+                            JsValue::from(src.0)
+                        }
+                    }
+
+                    impl #full_type_params std::convert::AsRef<JsValue> for #name #full_type_params {
+                        fn as_ref(&self) -> &JsValue {
+                            self.0.as_ref()
+                        }
+                    }
+
+                    impl #full_type_params wasm_bindgen::JsCast for #name #full_type_params {
+                        fn instanceof(val: &JsValue) -> bool {
+                            #internal_class_name::instanceof(val)
+                        }
+                        fn unchecked_from_js(val: JsValue) -> Self {
+                            #name(#internal_class_name::unchecked_from_js(val), #extra_args)
+                        }
+                        fn unchecked_from_js_ref(val: &JsValue) -> &Self {
+                            unsafe {
+                                &*(#internal_class_name::unchecked_from_js_ref(val) as *const #internal_class_name as *const Self)
+                            }
+                        }
+                    }
+
                     impl #full_type_params_deserializable #name #full_type_params {
                         #(#public_methods)*
                     }
@@ -1090,25 +1189,6 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                     panic!("Intersections must not be empty");
                 }
             }
-            /*TypeInfo::Mapped {
-                value_type: Box<TypeInfo>,
-            },
-            TypeInfo::LitNumber {
-                n: f64,
-            },
-            TypeInfo::LitString {
-                s: String,
-            },
-            TypeInfo::LitBoolean {
-                b: bool,
-            },
-            TypeInfo::Constructor {
-                params: Vec<Param>,
-                return_type: Box<TypeInfo>,
-            },
-            TypeInfo::Var {
-                type_info: Box<TypeInfo>,
-            },*/
             TargetEnrichedTypeInfo::NamespaceImport(NamespaceImport::All { src, .. }) => {
                 let ns = src.as_path().to_ns_path(*fs, type_name);
                 if ns.len() == 1 {
@@ -1271,48 +1351,13 @@ impl ToTokens for TargetEnrichedTypeInfo {
                     std::collections::HashMap<String, #value_type>
                 }
             }
-            TargetEnrichedTypeInfo::Func(Func {
-                params,
-                type_params: _,
-                return_type,
-                ..
-            }) => {
-                let param_toks: Vec<TokenStream2> = params
-                    .iter()
-                    .map(|p| {
-                        let typ = &p.type_info;
-
-                        if p.is_variadic {
-                            quote! {
-                                &[#typ]
-                            }
-                        } else {
-                            quote! {
-                                #typ
-                            }
-                        }
-                    })
-                    .collect();
+            TargetEnrichedTypeInfo::FuncGroup(fg) => {
+                let f = &fg.widened_fn;
 
                 quote! {
-                    fn(#(#param_toks),*) -> #return_type
+                    #f
                 }
             }
-            /*
-            TargetEnrichedTypeInfo::Constructor {
-                params: Vec<Param>,
-                return_type: Box<TypeInfo>,
-            },
-            TargetEnrichedTypeInfo::Class(Class {
-                members: HashMap<String, Member>,
-            }),
-            TargetEnrichedTypeInfo::Var {
-                type_info: Box<TypeInfo>,
-            },
-            TargetEnrichedTypeInfo::GenericType {
-                name: String,
-                constraint: Box<TypeInfo>,
-            },*/
             TargetEnrichedTypeInfo::NamespaceImport(_) => panic!("namespace import in type info"),
             _ => {
                 quote! {}
@@ -1323,6 +1368,41 @@ impl ToTokens for TargetEnrichedTypeInfo {
     }
 }
 
+fn func_to_tokens(params: &[Param], return_type: &TypeRef) -> TokenStream2 {
+    let param_toks: Vec<TokenStream2> = params
+        .iter()
+        .map(|p| {
+            let typ = &p.type_info;
+
+            if p.is_variadic {
+                quote! {
+                    &[#typ]
+                }
+            } else {
+                quote! {
+                    #typ
+                }
+            }
+        })
+        .collect();
+
+    quote! {
+        fn(#(#param_toks),*) -> #return_type
+    }
+}
+
+impl ToTokens for Func {
+    fn to_tokens(&self, toks: &mut TokenStream2) {
+        toks.append_all(func_to_tokens(&self.params, self.return_type.as_ref()));
+    }
+}
+
+impl ToTokens for Constructor<'_> {
+    fn to_tokens(&self, toks: &mut TokenStream2) {
+        toks.append_all(func_to_tokens(&self.ctor.params, self.class.as_ref()));
+    }
+}
+
 impl ToTokens for TypeRef {
     fn to_tokens(&self, toks: &mut TokenStream2) {
         let our_toks = {
@@ -1416,8 +1496,8 @@ struct OwnedTypeInfo<'a> {
 impl<'a> ToTokens for OwnedTypeInfo<'a> {
     fn to_tokens(&self, toks: &mut TokenStream2) {
         match self.type_info {
-            TargetEnrichedTypeInfo::Func(f) => {
-                let tr = OwnedTypeRef(Cow::Owned(f.clone().into()));
+            TargetEnrichedTypeInfo::FuncGroup(f) => {
+                let tr = OwnedTypeRef(Cow::Owned(f.widened_fn.clone().into()));
                 tr.to_tokens(toks);
             }
             TargetEnrichedTypeInfo::Ref(r) => {
@@ -1481,9 +1561,12 @@ fn render_deserialize_fn(
             let rendered_type = OwnedTypeRef(Cow::Borrowed(tr));
             (tr as &dyn HasFnPrototype, quote! { #rendered_type })
         }
-        TargetEnrichedTypeInfo::Func(f) => {
+        TargetEnrichedTypeInfo::FuncGroup(fg) => {
             let rendered_type = OwnedTypeInfo { type_info };
-            (f as &dyn HasFnPrototype, quote! { #rendered_type })
+            (
+                &fg.widened_fn as &dyn HasFnPrototype,
+                quote! { #rendered_type },
+            )
         }
         _ => {
             return None;
@@ -1529,7 +1612,7 @@ fn render_deserialize_fn(
         {
             let jsv: JsValue = ts_bindgen_rt::deserialize_as_jsvalue(deserializer)?;
             let #field_name: Option<&js_sys::Function> = wasm_bindgen::JsCast::dyn_ref(&jsv);
-            Ok(#field_name.map(|f| {
+            #field_name.map(|f| {
                 let f = f.clone();
                 std::rc::Rc::new(move |#(#params),*| {
                     #(#conversions);*
@@ -1540,7 +1623,7 @@ fn render_deserialize_fn(
                 }) as #rendered_type
             })
             .ok_or_else(|| ts_bindgen_rt::jsvalue_serde::Error::InvalidType("expected function".to_string()))
-            .map_err(serde::de::Error::custom)?)
+            .map_err(serde::de::Error::custom)
         }
     })
 }
@@ -1559,9 +1642,12 @@ fn render_serialize_fn(
             let rendered_type = OwnedTypeRef(Cow::Borrowed(tr));
             (tr as &dyn HasFnPrototype, quote! { #rendered_type })
         }
-        TargetEnrichedTypeInfo::Func(f) => {
+        TargetEnrichedTypeInfo::FuncGroup(f) => {
             let rendered_type = OwnedTypeInfo { type_info };
-            (f as &dyn HasFnPrototype, quote! { #rendered_type })
+            (
+                &f.widened_fn as &dyn HasFnPrototype,
+                quote! { #rendered_type },
+            )
         }
         _ => {
             return None;
