@@ -1,11 +1,13 @@
+use crate::codegen::funcs::Constructor;
 use crate::codegen::ns_path::ToNsPath;
 use crate::codegen::resolve_target_type::ResolveTargetType;
 use crate::fs::Fs;
 use crate::identifier::{
     make_identifier, to_camel_case_ident, to_ident, to_snake_case_ident, Identifier,
 };
-use crate::ir::{Builtin, TargetEnrichedTypeInfo, TypeIdent, TypeRef};
+use crate::ir::{Builtin, Func, TargetEnrichedTypeInfo, TypeIdent, TypeRef};
 use crate::mod_def::ToModPathIter;
+use quote::{quote, ToTokens};
 
 pub trait Named {
     /// Returns a tuple of the the js name as a string and the corresponding rust identifier
@@ -35,7 +37,6 @@ impl Named for Builtin {
             Builtin::PrimitiveBoolean => ("bool", to_ident("bool")),
             Builtin::PrimitiveBigInt => ("u64", to_ident("u64")),
             Builtin::PrimitiveString => ("String", to_ident("String")),
-            Builtin::PrimitiveSymbol => ("js_sys::Symbol", make_identifier!(js_sys::Symbol)),
             // TODO: is this correct?
             Builtin::PrimitiveVoid => ("()", to_ident("()")),
             Builtin::PrimitiveUndefined => (
@@ -45,7 +46,6 @@ impl Named for Builtin {
             Builtin::PrimitiveNull => {
                 ("ts_bindgen_rt::Null", make_identifier!(ts_bindgen_rt::Null))
             }
-            Builtin::Date => ("js_sys::Date", make_identifier!(js_sys::Date)),
             Builtin::LitNumber => Builtin::PrimitiveNumber.to_name(),
             Builtin::LitBoolean => Builtin::PrimitiveBoolean.to_name(),
             Builtin::LitString => Builtin::PrimitiveString.to_name(),
@@ -58,7 +58,7 @@ impl Named for Builtin {
             ),
             Builtin::Optional => ("Option", to_ident("Option")),
             Builtin::Variadic => ("", to_ident("")),
-            Builtin::Named(n) => (&n, Identifier::with_path(n)),
+            Builtin::Named(n) => (n, Identifier::with_path(n)),
         }
     }
 
@@ -227,7 +227,7 @@ impl SimpleNamed for TypeRef {
 
 fn casing_for_type<T: AsRef<str>>(typ: &TargetEnrichedTypeInfo) -> &dyn Fn(T) -> Identifier {
     match typ {
-        TargetEnrichedTypeInfo::Func(_) => &to_snake_case_ident as &dyn Fn(T) -> Identifier,
+        TargetEnrichedTypeInfo::FuncGroup(_) => &to_snake_case_ident as &dyn Fn(T) -> Identifier,
         TargetEnrichedTypeInfo::Ref(tr)
             if matches!(&tr.referent, TypeIdent::Builtin(Builtin::Fn)) =>
         {
@@ -236,3 +236,32 @@ fn casing_for_type<T: AsRef<str>>(typ: &TargetEnrichedTypeInfo) -> &dyn Fn(T) ->
         _ => &to_camel_case_ident as &dyn Fn(T) -> Identifier,
     }
 }
+
+pub fn type_name<T: ToTokens>(typ: &T) -> Identifier {
+    let t_str = quote! { #typ }
+        .to_string()
+        .replace("->", "To")
+        .replace(",", "And")
+        .replace("<", "Of")
+        .replace(">", "")
+        .replace("&", "")
+        .replace("[", "")
+        .replace("]", "");
+    to_camel_case_ident(t_str)
+}
+
+pub trait FnOverloadName {
+    fn overload_name(&self, fn_group_name: &Identifier) -> Identifier;
+}
+
+impl FnOverloadName for Func {
+    fn overload_name(&self, fn_group_name: &Identifier) -> Identifier {
+        fn_group_name.suffix_name(&format!("_{}", type_name(self)))
+    }
+}
+
+impl FnOverloadName for Constructor<'_> {
+    fn overload_name(&self, fn_group_name: &Identifier) -> Identifier {
+        fn_group_name.suffix_name(&format!("_{}", type_name(self)))
+    }
+}
