@@ -1,11 +1,14 @@
 use crate::codegen::resolve_target_type::ResolveTargetType;
 use crate::codegen::type_ref_like::{OwnedTypeRef, TypeRefLike};
 use crate::ir::{Builtin, TargetEnrichedTypeInfo, TypeIdent, TypeRef};
+use proc_macro2::TokenStream as TokenStream2;
+use quote::quote;
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub enum SerializationType {
     Raw,
     SerdeJson,
+    Array,
     Fn,
     JsValue,
 }
@@ -26,17 +29,18 @@ impl SerializationTypeGetter for TypeRef {
 impl SerializationTypeGetter for TargetEnrichedTypeInfo {
     fn serialization_type(&self) -> SerializationType {
         match self {
-            TargetEnrichedTypeInfo::Func(_) => SerializationType::Fn,
+            TargetEnrichedTypeInfo::FuncGroup(_) => SerializationType::Fn,
             TargetEnrichedTypeInfo::Class(_) => SerializationType::Raw,
             TargetEnrichedTypeInfo::Ref(t) => match &t.referent {
                 TypeIdent::Builtin(Builtin::Fn) => SerializationType::Fn,
-                TypeIdent::Builtin(Builtin::Array) => SerializationType::SerdeJson,
+                TypeIdent::Builtin(Builtin::Array) => SerializationType::Array,
                 TypeIdent::Builtin(
                     Builtin::PrimitiveAny | Builtin::PrimitiveObject | Builtin::Named(_),
                 ) => SerializationType::JsValue,
                 TypeIdent::Builtin(_) => SerializationType::Raw,
                 _ => SerializationType::SerdeJson,
             },
+            TargetEnrichedTypeInfo::Array { .. } => SerializationType::Array,
             _ => SerializationType::SerdeJson,
         }
     }
@@ -56,3 +60,36 @@ impl<'a> SerializationTypeGetter for TypeRefLike<'a> {
         }
     }
 }
+
+trait IsCopy {
+    fn is_copy(&self) -> bool;
+}
+
+impl IsCopy for TargetEnrichedTypeInfo {
+    fn is_copy(&self) -> bool {
+        match self {
+            TargetEnrichedTypeInfo::Ref(t) => {
+                matches!(
+                    &t.referent,
+                    TypeIdent::Builtin(
+                        Builtin::PrimitiveNumber
+                            | Builtin::PrimitiveBoolean
+                            | Builtin::LitNumber
+                            | Builtin::LitBoolean
+                    )
+                )
+            }
+            _ => false,
+        }
+    }
+}
+
+pub fn clone_item_of_type(item: TokenStream2, typ: &TargetEnrichedTypeInfo) -> TokenStream2 {
+    if typ.is_copy() {
+        item
+    } else {
+        quote! {
+            #item.clone()
+        }
+    }
+}
