@@ -1,8 +1,9 @@
 use crate::fs::Fs;
 use crate::identifier::Identifier;
 use crate::ir::flattened::{
-    Alias as FlattenedAlias, Class as FlattenedClass, Ctor as FlattenedCtor, Enum as FlattenedEnum,
-    EnumMember as FlattenedEnumMember, FlatType, FlattenedTypeInfo, Func as FlattenedFunc,
+    Alias as FlattenedAlias, Class as FlattenedClass, Ctor as FlattenedCtor,
+    CtorGroup as FlattenedCtorGroup, Enum as FlattenedEnum, EnumMember as FlattenedEnumMember,
+    FlatType, FlattenedTypeInfo, Func as FlattenedFunc, FuncGroup as FlattenedFuncGroup,
     Indexer as FlattenedIndexer, Interface as FlattenedInterface,
     Intersection as FlattenedIntersection, Member as FlattenedMember,
     NamespaceImport as FlattenedNamespaceImport, Param as FlattenedParam, Tuple as FlattenedTuple,
@@ -156,7 +157,7 @@ pub enum TargetEnrichedTypeInfo {
     Mapped {
         value_type: Box<TargetEnrichedTypeInfo>,
     },
-    Func(Func),
+    FuncGroup(FuncGroup),
     Constructor(Ctor),
     Class(Class),
     Var {
@@ -222,8 +223,8 @@ impl From<WithContext<FlattenedTypeInfo>> for TargetEnrichedTypeInfo {
             case_conv!(match value_type => FlattenedTypeInfo::Mapped, x) => {
                 case_conv!(value_type => TargetEnrichedTypeInfo::Mapped, x, ctx)
             }
-            case_conv!(match FlattenedTypeInfo::Func, x) => {
-                case_conv!(TargetEnrichedTypeInfo::Func, x, ctx)
+            case_conv!(match FlattenedTypeInfo::FuncGroup, x) => {
+                case_conv!(TargetEnrichedTypeInfo::FuncGroup, x, ctx)
             }
             case_conv!(match FlattenedTypeInfo::Constructor, x) => {
                 case_conv!(TargetEnrichedTypeInfo::Constructor, x, ctx)
@@ -246,8 +247,8 @@ impl From<WithContext<FlattenedTypeInfo>> for TargetEnrichedTypeInfo {
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub enum Member {
-    Constructor(Ctor),
-    Method(Func),
+    Constructor(CtorGroup),
+    Method(FuncGroup),
     Property(TypeRef),
 }
 
@@ -279,6 +280,17 @@ pub struct Class {
     pub context: Context,
 }
 
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct CtorGroup {
+    pub overloads: Vec<Ctor>,
+    pub context: Context,
+}
+
+from_struct!(
+    FlattenedCtorGroup => CtorGroup;
+    overloads => [],
+);
+
 from_struct!(
     FlattenedClass => Class;
     super_class => Option,
@@ -326,6 +338,19 @@ from_struct!(
     default_type_arg => Option,
 );
 
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct FuncGroup {
+    pub overloads: Vec<Func>,
+    pub widened_fn: Func,
+    pub context: Context,
+}
+
+from_struct!(
+    FlattenedFuncGroup => FuncGroup;
+    overloads => [],
+    widened_fn => .,
+);
+
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Func {
     pub type_params: Vec<(String, TypeParamConfig)>,
